/*
 * am.c
 *
 * Anthony's Make
 *
 * Copyright 1993, 1994 by Anthony Howe.  All rights reserved.  No warranty.
 *
 *
 * SYNOPSIS
 *
 * am <makefile> [targets...]
 *
 *
 * REFERENCES
 *
 * [Mil87]	Webb Miller, "A Software Tools Sampler", Prentic Hall, 87
 *		ISBN 0-13-822305-X, chaper 2
 *
 * [POSIX]	POSIX.2 MAKE
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#ifdef __STDC__
#define _(x)			x
#else
#define _(x)			()
#endif

#ifndef RULES
#define RULES			500
#endif

#ifndef WORK_AREA_SIZE
#define WORK_AREA_SIZE		32767
#endif

#ifndef MACRO_NAME_LENGTH
#define MACRO_NAME_LENGTH	128
#endif

/*
 * Return start and length of next element in list.
 *
 * Initially p is a string pointer representing a list where each 
 * element in the list is sepearated by white space.
 */
#define next(p, i)	(p += strspn(p, white), i = strcspn(p, white))

typedef struct t_rule {
	char *tar;		/* List of targets. */
	char *dep;		/* List of dependencies */
	char *cmd;		/* First command in receipe. */
} t_rule;

t_rule rule[RULES];

char *white = "\n\t ";
char buffer[WORK_AREA_SIZE], *unused;
unsigned long line_no;

time_t mtime _((char *));
int target _((char *));
void update _((char *));

void load _((FILE *));
char *getline _((FILE *));
char *getmacro _((FILE *));
void fatal _((char *, ...));

int
main(argc, argv)
int argc;
char **argv;
{
	FILE *fp;
	struct stat s;

	if (argc < 2) {
		fprintf(stderr, "usage:  am <makefile> [targets...]\n");
		return 2;
	}

	if (stat(*++argv, &s) != 0 || (fp = fopen(*argv, "r")) == NULL)
		return 1;

	load(fp);
	(void) fclose(fp);

	if (argc == 2) {
		update(rule[1].tar);
	} else {
		for (--argc, ++argv; 1 < argc; --argc, ++argv)
			update(*argv);
	}

	return 0;
}

/*
 * Return the modification time for file.
 * If file does not exist then return zero.
 */
time_t
mtime(file)
char *file;
{
	size_t len;
	int value, ch;
	struct stat s;

	next(file, len);
	ch = file[len];
	file[len] = '\0'; 
	value = stat(file, &s);
	file[len] = ch;

	return value < 0 ? 0 : s.st_mtime;
}

/*
 * Find first occurence of target file.  All subsequent occurences
 * are ignored.  Return the rule number, zero (0) is the empty
 * rule and indicates no target found.
 */
int
target(file)
char *file;
{
	char *tar;
	size_t tlen, flen, k = 1;

	for (next(file, flen); (tar = rule[k].tar) != NULL; ++k) {
		for (; next(tar, tlen) != 0; tar += tlen) {
			if (flen == tlen && strncmp(file, tar, flen) == 0)
				return k;
		}
	}

	return 0;
}

/*
 * Try to update a file's dependencies, then the file itself.
 * If the file does not exist, has no dependencies, and no 
 * known commands, then exit with an error.
 */
void
update(file)
char *file;
{
	char *dep, *cmd;
	time_t last_touched = mtime(file);
	size_t i, j = last_touched == 0, k = target(file);

	dep = rule[k].dep;
	cmd = rule[k].cmd;

	/* If file does not exist, has an empty dependency list, 
	 * and an improper receipe, then we cannot build file.
	 */
	if (next(dep, i) == 0 && cmd != NULL && *cmd != '\t' && j)
		fatal("Do not know how to make \"%s\".", file);

	/* Try and update all out-of-date dependencies first. */
	for ( ; 0 < i; dep += i, next(dep, i)) {
		update(dep);
		if (last_touched < mtime(dep))
			j = 1;
	}

	if (j && cmd != NULL && *cmd == '\t') {
		/* Some systems do not record a file's modification times
		 * with enough precision.  This sleep() guarantees that
		 * each file has a unique modification time.
		 */
		sleep(1);

		/* Execute each command in receipe.  A receipe consists of
		 * command lines, each of which is prefixed by a tab.  The
		 * first line that does not have an initial tab ends the 
		 * receipe. 
		 */
		for (; *cmd++ == '\t'; cmd += strlen(cmd) + 1) {
			/* Look for command line modifier flags. */
			for (i = j = 1; ; ++cmd) {
				switch (*cmd) {
				case '@':
					/* Be silent. */
					i = 0;
					continue;
				case '-':
					/* Ignore errors. */
					j = 0;
					continue;
				case '+':
					/* Always execute. */
					continue;
				}
				break;
			}

			if (i)
				puts(cmd);

			if (system(cmd) != 0 && j)
				fatal("Error making \"%s\".", file);
		}

		/* Mark rule's receipe as having been done. */
		rule[k].cmd = NULL;
	}
}

/*
 * Read in and parse makefile.
 */
void
load(fp)
FILE *fp;
{
	int count;
	char *line, *value;

	/* The empty rule. */
	rule[0].tar = rule[0].dep = rule[0].cmd = "";

	unused = buffer;

	for (line_no = count = 1; (line = getline(fp)) != NULL; ) {
		if (rule[count].tar != NULL) {
			/* Line following a target should be the
			 * first line of a command receipe, provide
			 * it begins with a <tab>. 
			 */
			rule[count++].cmd = *line == '\t' ? line : NULL;
		} 

		if (*line == '\t') {
			/* Keep receipe lines. */
			;
		} else if (*line == '#' || *line == '\n') {
			/* Throw away comment or blank line. */
			unused = line;
		} else if ((value = strchr(line, ':')) != NULL) {
			/* Add target rule to table. */
			if (RULES <= count)
				fatal("#Too many rules, maximum %d.", RULES);

			rule[count].tar = line;
			rule[count].dep = value + 1;
			*value = '\0';
		} else if ((value = strchr(line, '=')) != NULL) {
			/* Add macro to environment. */
			char *macro, *emacro;

			/* Strip whitespace between the macro name,
			 * equal-sign, and the macro text.
			 */
			macro = line + strspn(line, white);
			emacro = macro + strcspn(macro, " \t=");
			*emacro++ = '=';

			value += strspn(++value, white);
			(void) strcpy(emacro, value);

			if (putenv(macro) == -1)
				fatal("#Out of environment space for macros.");
		}
	}
}

/*
 * Read in a complete line from the makefile into the
 * workspace.  Handle escaped newlines and macro
 * expansions.  Return a pointer to the start of the
 * line within the workspace.
 */
char *
getline(fp)
FILE *fp;
{
	int ch;
	char *start, *value;

	for (start = unused; unused - buffer < sizeof buffer; ) {
		switch (ch = fgetc(fp)) {
		case EOF:
			return NULL;
		case '\n':
			++line_no;
		case '\r':
			if (unused[-1] == '\\') {
				/* Escaped newline, compress whitespace. */
				do
					ch = fgetc(fp);
				while (isspace(ch));
				(void) ungetc(ch, fp);
				unused[-1] = ' ';
				break;
			}
			*unused++ = '\0';
			return start;
		case '$':
			if ((value = getmacro(fp)) != NULL) {
				size_t len = strlen(value);

				if (sizeof buffer <= (unused-buffer) + len + 1)
					fatal("#Out of workspace.");

				(void) strcpy(unused, value);
				unused += len;
			}
			break;
		default:
			*unused++ = ch;
		}
	}

	fatal("#Out of workspace.");
}

/*
 * Read in a macro name and find it's value.  Handle $$, 
 * $m, $(macro), ${macro}, as well as nested macros.
 * Return a pointer to a static string.
 */
char *
getmacro(fp)
FILE *fp;
{
	int ch, ebracket, mlen, vlen;
	char *macro, *value, single[2];
	static char *dollar_sign = "$";

	if ((ch = fgetc(fp)) == EOF)
		fatal("#Premature end-of-file.");

	if (ch == '$')
		return dollar_sign;

	if (isalpha(ch)) {
		single[0] = ch;
		single[1] = '\0';
		return getenv(single);
	}

	if (ch == '(')
		ebracket = ')';
	else if (ch == '{')
		ebracket = '}';
	else
		fatal("#Badly formed macro.");

	if ((macro = (char *) malloc(MACRO_NAME_LENGTH)) == NULL)
		return NULL;

	for (mlen = 0; mlen < MACRO_NAME_LENGTH; ) {
		switch (ch = fgetc(fp)) {
		case EOF:
			fatal("#Premature end-of-file.");
		case '$':
			if ((value = getmacro(fp)) != NULL && *value != '\0') {
				vlen = (int) strlen(value);
				if (MACRO_NAME_LENGTH <= mlen + vlen)
					fatal("#Macro name too long.");

				(void) strcpy(&macro[mlen], value);
				mlen += vlen;
			}
			break;
		case '\n':
			++line_no;
			/* fall */
		default:
			if (ch == ebracket) {
				macro[mlen] = '\0';
				value = getenv(macro);
				free(macro);
				return value;
			}

			macro[mlen++] = ch;
		}
	}

	fatal("#Macro name too long.");
}
	
/*
 * Report line number and error, then exit program.
 */
void
#ifdef __STDC__
fatal(char *fmt, ...)
#else
fatal(fmt)
char *fmt;
#endif /* __STDC__ */
{
	va_list args;

	fputs("am: ", stderr);

	if (*fmt == '#') {
		fprintf(stderr, "line %lu: ", line_no);
		++fmt;
	}

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	fputc('\n', stderr);
	exit(1);
}

